@@ -22,7 +22,11 @@ std::string Constraint::DescriptionString() const {
         case Type::EQ_LEN_PT_LINE_D:    s = C_("constr-name", "eq-length-and-pt-ln-dist"); break;
         case Type::EQ_PT_LN_DISTANCES:  s = C_("constr-name", "eq-pt-line-distances"); break;
         case Type::LENGTH_RATIO:        s = C_("constr-name", "length-ratio"); break;
+        case Type::ARC_ARC_LEN_RATIO:   s = C_("constr-name", "arc-arc-length-ratio"); break;
+        case Type::ARC_LINE_LEN_RATIO:  s = C_("constr-name", "arc-line-length-ratio"); break;
         case Type::LENGTH_DIFFERENCE:   s = C_("constr-name", "length-difference"); break;
+        case Type::ARC_ARC_DIFFERENCE:  s = C_("constr-name", "arc-arc-len-difference"); break;
+        case Type::ARC_LINE_DIFFERENCE: s = C_("constr-name", "arc-line-len-difference"); break;
         case Type::SYMMETRIC:           s = C_("constr-name", "symmetric"); break;
         case Type::SYMMETRIC_HORIZ:     s = C_("constr-name", "symmetric-h"); break;
         case Type::SYMMETRIC_VERT:      s = C_("constr-name", "symmetric-v"); break;
@@ -127,6 +131,69 @@ hConstraint Constraint::ConstrainCoincident(hEntity ptA, hEntity ptB) {
         Entity::NO_ENTITY, Entity::NO_ENTITY, /*other=*/false, /*other2=*/false);
 }
 
+bool Constraint::ConstrainArcLineTangent(Constraint *c, Entity *line, Entity *arc) {
+    Vector l0 = SK.GetEntity(line->point[0])->PointGetNum(),
+           l1 = SK.GetEntity(line->point[1])->PointGetNum();
+    Vector a1 = SK.GetEntity(arc->point[1])->PointGetNum(),
+           a2 = SK.GetEntity(arc->point[2])->PointGetNum();
+    if(l0.Equals(a1) || l1.Equals(a1)) {
+        c->other = false;
+    } else if(l0.Equals(a2) || l1.Equals(a2)) {
+        c->other = true;
+    } else {
+        Error(_("The tangent arc and line segment must share an "
+                "endpoint. Constrain them with Constrain -> "
+                "On Point before constraining tangent."));
+        return false;
+    }
+    return true;
+}
+
+bool Constraint::ConstrainCubicLineTangent(Constraint *c, Entity *line, Entity *cubic) {
+    Vector l0 = SK.GetEntity(line->point[0])->PointGetNum(),
+           l1 = SK.GetEntity(line->point[1])->PointGetNum();
+    Vector as = cubic->CubicGetStartNum(),
+           af = cubic->CubicGetFinishNum();
+
+    if(l0.Equals(as) || l1.Equals(as)) {
+        c->other = false;
+    } else if(l0.Equals(af) || l1.Equals(af)) {
+        c->other = true;
+    } else {
+        Error(_("The tangent cubic and line segment must share an "
+                "endpoint. Constrain them with Constrain -> "
+                "On Point before constraining tangent."));
+        return false;
+    }
+    return true;
+}
+
+bool Constraint::ConstrainCurveCurveTangent(Constraint *c, Entity *eA, Entity *eB) {
+    Vector as = eA->EndpointStart(),
+           af = eA->EndpointFinish(),
+           bs = eB->EndpointStart(),
+           bf = eB->EndpointFinish();
+    if(as.Equals(bs)) {
+        c->other = false;
+        c->other2 = false;
+    } else if(as.Equals(bf)) {
+        c->other = false;
+        c->other2 = true;
+    } else if(af.Equals(bs)) {
+        c->other = true;
+        c->other2 = false;
+    } else if(af.Equals(bf)) {
+        c->other = true;
+        c->other2 = true;
+    } else {
+        Error(_("The curves must share an endpoint. Constrain them "
+                "with Constrain -> On Point before constraining "
+                "tangent."));
+        return false;
+    }
+    return true;
+}
+
 void Constraint::MenuConstrain(Command id) {
     Constraint c = {};
     c.group = SS.GW.activeGroup;
@@ -321,10 +388,27 @@ void Constraint::MenuConstrain(Command id) {
                 c.type = Type::LENGTH_RATIO;
                 c.entityA = gs.entity[0];
                 c.entityB = gs.entity[1];
+            }
+            else if(gs.arcs == 2 && gs.n == 2) {
+                c.type = Type::ARC_ARC_LEN_RATIO;
+                c.entityA = gs.entity[0];
+                c.entityB = gs.entity[1];
+            }
+            else if(gs.lineSegments == 1 && gs.arcs == 1 && gs.n == 2) {
+                c.type = Type::ARC_LINE_LEN_RATIO;
+                if(SK.GetEntity(gs.entity[0])->type == Entity::Type::ARC_OF_CIRCLE) {
+                    c.entityA = gs.entity[1];
+                    c.entityB = gs.entity[0];
+                } else {
+                    c.entityA = gs.entity[0];
+                    c.entityB = gs.entity[1];
+                }
             } else {
                 Error(_("Bad selection for length ratio constraint. This "
                         "constraint can apply to:\n\n"
-                        "    * two line segments\n"));
+                        "    * two line segments\n"
+                        "    * two arcs\n"
+                        "    * one arc and one line segment\n"));
                 return;
             }
 
@@ -338,10 +422,27 @@ void Constraint::MenuConstrain(Command id) {
                 c.type = Type::LENGTH_DIFFERENCE;
                 c.entityA = gs.entity[0];
                 c.entityB = gs.entity[1];
+            }
+            else if(gs.arcs == 2 && gs.n == 2) {
+                c.type = Type::ARC_ARC_DIFFERENCE;
+                c.entityA = gs.entity[0];
+                c.entityB = gs.entity[1];
+            }
+            else if(gs.lineSegments == 1 && gs.arcs == 1 && gs.n == 2) {
+                c.type = Type::ARC_LINE_DIFFERENCE;
+                if(SK.GetEntity(gs.entity[0])->type == Entity::Type::ARC_OF_CIRCLE) {
+                    c.entityA = gs.entity[1];
+                    c.entityB = gs.entity[0];
+                } else {
+                    c.entityA = gs.entity[0];
+                    c.entityB = gs.entity[1];
+                }
             } else {
                 Error(_("Bad selection for length difference constraint. This "
                         "constraint can apply to:\n\n"
-                        "    * two line segments\n"));
+                        "    * two line segments\n"
+                        "    * two arcs\n"
+                        "    * one arc and one line segment\n"));
                 return;
             }
 
@@ -357,8 +458,11 @@ void Constraint::MenuConstrain(Command id) {
                 c.ptA = gs.point[0];
 
                 // If a point is at-midpoint, then no reason to also constrain
-                // it on-line; so auto-remove that.
+                // it on-line; so auto-remove that.  Handle as one undo group.
+                SS.UndoRemember();
                 DeleteAllConstraintsFor(Type::PT_ON_LINE, c.entityA, c.ptA);
+                AddConstraint(&c, /*rememberForUndo=*/false);
+                break;
             } else if(gs.lineSegments == 1 && gs.workplanes == 1 && gs.n == 2) {
                 c.type = Type::AT_MIDPOINT;
                 int i = SK.GetEntity(gs.entity[0])->IsWorkplane() ? 1 : 0;
@@ -433,6 +537,7 @@ void Constraint::MenuConstrain(Command id) {
                             "(symmetric about workplane)\n"));
                 return;
             }
+            // We may remove constraints so remember manually
             if(c.entityA == Entity::NO_ENTITY) {
                 // Horizontal / vertical symmetry, implicit symmetry plane
                 // normal to the workplane
@@ -454,10 +559,14 @@ void Constraint::MenuConstrain(Command id) {
                 if(gs.lineSegments == 1) {
                     // If this line segment is already constrained horiz or
                     // vert, then auto-remove that redundant constraint.
+                    // Handle as one undo group.
+                    SS.UndoRemember();
                     DeleteAllConstraintsFor(Type::HORIZONTAL, (gs.entity[0]),
                         Entity::NO_ENTITY);
                     DeleteAllConstraintsFor(Type::VERTICAL, (gs.entity[0]),
                         Entity::NO_ENTITY);
+                    AddConstraint(&c, /*rememberForUndo=*/false);
+                    break;
                 }
             }
             AddConstraint(&c);
@@ -558,6 +667,7 @@ void Constraint::MenuConstrain(Command id) {
             if(gs.constraints == 1 && gs.n == 0) {
                 Constraint *c = SK.GetConstraint(gs.constraint[0]);
                 if(c->HasLabel() && c->type != Type::COMMENT) {
+                    SS.UndoRemember();
                     (c->reference) = !(c->reference);
                     SS.MarkGroupDirty(c->group, /*onlyThis=*/true);
                     break;
@@ -611,53 +721,33 @@ void Constraint::MenuConstrain(Command id) {
         }
 
         case Command::PARALLEL:
-            if(gs.vectors == 2 && gs.n == 2) {
+            if(gs.faces == 2 && gs.n == 2) {
+                c.type = Type::PARALLEL;
+                c.entityA = gs.face[0];
+                c.entityB = gs.face[1];            
+            } else if(gs.vectors == 2 && gs.n == 2) {
                 c.type = Type::PARALLEL;
                 c.entityA = gs.vector[0];
                 c.entityB = gs.vector[1];
             } else if(gs.lineSegments == 1 && gs.arcs == 1 && gs.n == 2) {
-                Entity *line = SK.GetEntity(gs.entity[0]);
-                Entity *arc  = SK.GetEntity(gs.entity[1]);
+                Entity *line = SK.GetEntity(gs.entity[0]),
+                       *arc  = SK.GetEntity(gs.entity[1]);
                 if(line->type == Entity::Type::ARC_OF_CIRCLE) {
                     swap(line, arc);
                 }
-                Vector l0 = SK.GetEntity(line->point[0])->PointGetNum(),
-                       l1 = SK.GetEntity(line->point[1])->PointGetNum();
-                Vector a1 = SK.GetEntity(arc->point[1])->PointGetNum(),
-                       a2 = SK.GetEntity(arc->point[2])->PointGetNum();
-
-                if(l0.Equals(a1) || l1.Equals(a1)) {
-                    c.other = false;
-                } else if(l0.Equals(a2) || l1.Equals(a2)) {
-                    c.other = true;
-                } else {
-                    Error(_("The tangent arc and line segment must share an "
-                            "endpoint. Constrain them with Constrain -> "
-                            "On Point before constraining tangent."));
+                if(!ConstrainArcLineTangent(&c, line, arc)) {
                     return;
                 }
                 c.type = Type::ARC_LINE_TANGENT;
                 c.entityA = arc->h;
                 c.entityB = line->h;
             } else if(gs.lineSegments == 1 && gs.cubics == 1 && gs.n == 2) {
-                Entity *line  = SK.GetEntity(gs.entity[0]);
-                Entity *cubic = SK.GetEntity(gs.entity[1]);
+                Entity *line  = SK.GetEntity(gs.entity[0]),
+                       *cubic = SK.GetEntity(gs.entity[1]);
                 if(line->type == Entity::Type::CUBIC) {
                     swap(line, cubic);
                 }
-                Vector l0 = SK.GetEntity(line->point[0])->PointGetNum(),
-                       l1 = SK.GetEntity(line->point[1])->PointGetNum();
-                Vector as = cubic->CubicGetStartNum(),
-                       af = cubic->CubicGetFinishNum();
-
-                if(l0.Equals(as) || l1.Equals(as)) {
-                    c.other = false;
-                } else if(l0.Equals(af) || l1.Equals(af)) {
-                    c.other = true;
-                } else {
-                    Error(_("The tangent cubic and line segment must share an "
-                            "endpoint. Constrain them with Constrain -> "
-                            "On Point before constraining tangent."));
+                if(!ConstrainCubicLineTangent(&c, line, cubic)) {
                     return;
                 }
                 c.type = Type::CUBIC_LINE_TANGENT;
@@ -670,22 +760,7 @@ void Constraint::MenuConstrain(Command id) {
                 }
                 Entity *eA = SK.GetEntity(gs.entity[0]),
                        *eB = SK.GetEntity(gs.entity[1]);
-                Vector as = eA->EndpointStart(),
-                       af = eA->EndpointFinish(),
-                       bs = eB->EndpointStart(),
-                       bf = eB->EndpointFinish();
-                if(as.Equals(bs)) {
-                    c.other = false; c.other2 = false;
-                } else if(as.Equals(bf)) {
-                    c.other = false; c.other2 = true;
-                } else if(af.Equals(bs)) {
-                    c.other = true; c.other2 = false;
-                } else if(af.Equals(bf)) {
-                    c.other = true; c.other2 = true;
-                } else {
-                    Error(_("The curves must share an endpoint. Constrain them "
-                            "with Constrain -> On Point before constraining "
-                            "tangent."));
+                if(!ConstrainCurveCurveTangent(&c, eA, eB)) {
                     return;
                 }
                 c.type = Type::CURVE_CURVE_TANGENT;
@@ -694,6 +769,7 @@ void Constraint::MenuConstrain(Command id) {
             } else {
                 Error(_("Bad selection for parallel / tangent constraint. This "
                         "constraint can apply to:\n\n"
+                        "    * two faces\n"
                         "    * two line segments (parallel)\n"
                         "    * a line segment and a normal (parallel)\n"
                         "    * two normals (parallel)\n"
@@ -705,13 +781,18 @@ void Constraint::MenuConstrain(Command id) {
             break;
 
         case Command::PERPENDICULAR:
-            if(gs.vectors == 2 && gs.n == 2) {
+            if(gs.faces == 2 && gs.n == 2) {
+                c.type = Type::PERPENDICULAR;
+                c.entityA = gs.face[0];
+                c.entityB = gs.face[1];            
+            } else if(gs.vectors == 2 && gs.n == 2) {
                 c.type = Type::PERPENDICULAR;
                 c.entityA = gs.vector[0];
                 c.entityB = gs.vector[1];
             } else {
                 Error(_("Bad selection for perpendicular constraint. This "
                         "constraint can apply to:\n\n"
+                        "    * two faces\n"
                         "    * two line segments\n"
                         "    * a line segment and a normal\n"
                         "    * two normals\n"));
@@ -734,10 +815,19 @@ void Constraint::MenuConstrain(Command id) {
             break;
 
         case Command::COMMENT:
-            SS.GW.pending.operation = GraphicsWindow::Pending::COMMAND;
-            SS.GW.pending.command = Command::COMMENT;
-            SS.GW.pending.description = _("click center of comment text");
-            SS.ScheduleShowTW();
+            if(gs.points == 1 && gs.n == 1) {
+                c.type = Type::COMMENT;
+                c.ptA = gs.point[0];
+                c.group       = SS.GW.activeGroup;
+                c.workplane   = SS.GW.ActiveWorkplane();
+                c.comment     = _("NEW COMMENT -- DOUBLE-CLICK TO EDIT");
+                AddConstraint(&c);
+            } else {
+                SS.GW.pending.operation = GraphicsWindow::Pending::COMMAND;
+                SS.GW.pending.command = Command::COMMENT;
+                SS.GW.pending.description = _("click center of comment text");
+                SS.ScheduleShowTW();
+            }
             break;
 
         default: ssassert(false, "Unexpected menu ID");
